/* Statement-expressions parsing for Ksi.
   Copyright (C) 2001 Michal Moskal (malekith@pld.org.pl)

This file is part of GCC.

GCC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GCC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */


#include "ksi.h"

/* bind(EXP) -> scope */
tree compile_bind(tree *type)
{
	struct term_state ctx;
	tree exp, block, vars;
	
	pushlevel(0);

	init_term(&ctx, kw_bind);
	exp = compile(get_keyword(&ctx), type);
	fini_term(&ctx);
	
	vars = get_vars();
	block = poplevel(1, 0, 0);

	return build(BIND_EXPR, *type, vars, exp, block);
}

/* return(EXP) */
/* Note: return(const(type(int, 32), 10)) won't return 10, */
/* return(modify(result(), const(type(int, 32), 10))) will. */
tree compile_return(tree *type)
{
	struct term_state ctx;
	tree exp;

	init_term(&ctx, kw_return);
	exp = compile(get_keyword(&ctx), type);
	fini_term(&ctx);

	*type = void_type_node;
	
	exp = build1(RETURN_EXPR, *type, exp);

	return exp;
}

/* loop(EXP) */
tree compile_loop(tree *type)
{
	struct term_state ctx;
	tree exp;

	init_term(&ctx, kw_loop);
	exp = compile(get_keyword(&ctx), type);
	fini_term(&ctx);

	*type = void_type_node;
	
	exp = build1(LOOP_EXPR, *type, exp);

	return exp;
}

/* exit(COND) */
/* exit(const(type(bool), 1)) == break */
tree compile_exit_loop(tree *type)
{
	struct term_state ctx;
	tree exp;

	init_term(&ctx, kw_exit_loop);
	exp = compile(get_keyword(&ctx), type);
	fini_term(&ctx);

	if (TREE_CODE(*type) != BOOLEAN_TYPE)
		fatal_error("condition for (exit_loop ...) has non-bool type");

	*type = void_type_node;
	
	exp = build1(EXIT_EXPR, *type, exp);

	return exp;
}

#ifdef LABELS
/* labels are somewhat diffrent then other identifiers, */
/* they don't disappear until end of function. */

static tree labels_chain;

static tree find_label(tree id)
{
	tree ptr;

	for (ptr = labels_chain; ptr; ptr = TREE_CHAIN(ptr))
		if (DECL_NAME(ptr) == id) 
			return ptr;
	
	return NULL;
}

/* label(name) */
tree compile_label(tree *type)
{
	struct arg a;
	struct term_state ctx;
	tree decl, name, exp;

	init_term(&ctx, kw_label);
	get_term_arg(&ctx, arg_id, &a);
	name = get_identifier(a.u.val_id);
	free_arg(&a);

	fini_term(&ctx);

	if ((decl = find_label(name))) {
		if (TREE_LANG_FLAG_0(decl))	/* has been defined */
			fatal_error("multiple definition of label '%s'",
				    IDENTIFIER_POINTER(name));
	} else {
		decl = build_decl(LABEL_DECL, name, void_type_node);
		DECL_CONTEXT(decl) = current_function_decl;
		TREE_CHAIN(decl) = labels_chain;
		labels_chain = decl;
	}
	
	TREE_LANG_FLAG_0(decl) = 1;

	*type = void_type_node;
	
	exp = build1(LABEL_EXPR, *type, decl);
	/* this couses do_abort() to be called... */
	TREE_SIDE_EFFECTS(exp) = 1;

	return exp;
}

/* goto(name) */
tree compile_goto(tree *type)
{
	struct arg a;
	struct term_state ctx;
	tree id, decl, exp;

	init_term(&ctx, kw_goto);
	
	get_term_arg(&ctx, arg_id, &a);
	id = get_identifier(a.u.val_id);
	free_arg(&a);

	fini_term(&ctx);

	if ((decl = find_label(id)) == NULL) {
		decl = build_decl(LABEL_DECL, id, void_type_node);
		DECL_CONTEXT(decl) = current_function_decl;
		TREE_LANG_FLAG_0(decl) = 0;
		
		TREE_CHAIN(decl) = labels_chain;
		labels_chain = decl;
	}

	TREE_LANG_FLAG_1(decl) = 1;
	
	*type = void_type_node;
	
	exp = build1(GOTO_EXPR, *type, decl);

	TREE_SIDE_EFFECTS(exp) = 1;

	return exp;
}
#endif

struct block {
	struct block *chain;
	tree label;
	tree block;
	tree type;
};

static struct block *block_chain;

void fini_labels()
{
#ifdef LABELS
	tree ptr;

	for (ptr = labels_chain; ptr; ptr = TREE_CHAIN(ptr))
		if (!TREE_LANG_FLAG_0(ptr))
			fatal_error("label '%s' used but not defined",
				    IDENTIFIER_POINTER(DECL_NAME(ptr)));
		else if (!TREE_LANG_FLAG_1(ptr))
			warning("label '%s' defined but not used",
				IDENTIFIER_POINTER(DECL_NAME(ptr)));
	
	labels_chain = NULL;
#endif
	if (block_chain != NULL)
		ice();
}

tree compile_exit(tree *type)
{
	struct arg a;
	struct term_state ctx;
	struct block *b;
	tree exp, id;
	
	init_term(&ctx, kw_block);
	
	get_term_arg(&ctx, arg_id, &a);
	id = get_identifier(a.u.val_id);
	free_arg(&a);

	if (optional_term_arg_present(&ctx)) {
		exp = compile(get_keyword(&ctx), type);
	} else {
		*type = void_type_node;
		exp = build1(NOP_EXPR, *type, integer_zero_node);
	}
	
	fini_term(&ctx);

	for (b = block_chain; b; b = b->chain)
		if (DECL_NAME(b->label) == id)
			break;
	
	if (b == NULL)
		fatal_error("(exit ...) from undefined block '%s'",
			    IDENTIFIER_POINTER(id));
	
	same_types(kw_exit, *type, b->type);

	*type = void_type_node;

	exp = build(EXIT_BLOCK_EXPR, *type, b->block, 
		    /* this worksaround not implemented return_value */
		    /* in EXIT_BLOCK_EXPR */
		    b->type == void_type_node ? NULL_TREE : exp);
	TREE_SIDE_EFFECTS(exp) = 1;

	return exp;
}

tree compile_block(tree *type)
{
	struct arg a;
	struct term_state ctx;
	struct block b;
	keyword_t kw;
	tree decl, id, exp;
	
	init_term(&ctx, kw_block);
	
	get_term_arg(&ctx, arg_id, &a);
	id = get_identifier(a.u.val_id);
	free_arg(&a);

	decl = build_decl(LABEL_DECL, id, void_type_node);
	DECL_CONTEXT(decl) = current_function_decl;
	
	b.label = decl;
	b.chain = block_chain;
	block_chain = &b;
	
	kw = get_keyword(&ctx);
	if (kw == kw_type) {
		b.type = parse_type();
		kw = get_keyword(&ctx);
	} else
		b.type = void_type_node;
		
	b.block = build(LABELED_BLOCK_EXPR, b.type, decl, NULL_TREE);
	TREE_SIDE_EFFECTS(b.block) = 1;
	
	exp = compile(kw, type);
	
	fini_term(&ctx);
	
	/* don't check in case void is expected */
	if (b.type != void_type_node)
		same_types(kw_block, b.type, *type);
		
	/* pop it off */
	if (block_chain != &b)
		ice();
	block_chain = block_chain->chain;
	
	*type = b.type;

	LABELED_BLOCK_BODY(b.block) = exp;

	return b.block;
}
